import logger from '@wdio/logger';
import { commandCallStructure, isValidParameter, getArgumentType } from '@wdio/utils';
import RequestFactory from './request/factory.js';
const log = logger('webdriver');
const BIDI_COMMANDS = ['send', 'sendAsync'];
export default function (method, endpointUri, commandInfo, doubleEncodeVariables = false) {
    const { command, ref, parameters, variables = [], isHubCommand = false } = commandInfo;
    return async function protocolCommand(...args) {
        let endpoint = endpointUri; // clone endpointUri in case we change it
        const commandParams = [...variables.map((v) => Object.assign(v, {
                /**
                 * url variables are:
                 */
                required: true,
                type: 'string' // have to be always type of string
            })), ...parameters];
        const commandUsage = `${command}(${commandParams.map((p) => p.name).join(', ')})`;
        const moreInfo = `\n\nFor more info see ${ref}\n`;
        const body = {};
        /**
         * parameter check
         */
        const minAllowedParams = commandParams.filter((param) => param.required).length;
        if (args.length < minAllowedParams || args.length > commandParams.length) {
            const parameterDescription = commandParams.length
                ? `\n\nProperty Description:\n${commandParams.map((p) => `  "${p.name}" (${p.type}): ${p.description}`).join('\n')}`
                : '';
            throw new Error(`Wrong parameters applied for ${command}\n` +
                `Usage: ${commandUsage}` +
                parameterDescription +
                moreInfo);
        }
        /**
         * parameter type check
         */
        for (const [it, arg] of Object.entries(args)) {
            const i = parseInt(it, 10);
            const commandParam = commandParams[i];
            if (!isValidParameter(arg, commandParam.type)) {
                /**
                 * ignore if argument is not required
                 */
                if (typeof arg === 'undefined' && !commandParam.required) {
                    continue;
                }
                const actual = commandParam.type.endsWith('[]')
                    ? `(${(Array.isArray(arg) ? arg : [arg]).map((a) => getArgumentType(a))})[]`
                    : getArgumentType(arg);
                throw new Error(`Malformed type for "${commandParam.name}" parameter of command ${command}\n` +
                    `Expected: ${commandParam.type}\n` +
                    `Actual: ${actual}` +
                    moreInfo);
            }
            /**
             * inject url variables
             */
            if (i < variables.length) {
                const encodedArg = doubleEncodeVariables ? encodeURIComponent(encodeURIComponent(arg)) : encodeURIComponent(arg);
                endpoint = endpoint.replace(`:${commandParams[i].name}`, encodedArg);
                continue;
            }
            /**
             * rest of args are part of body payload
             */
            body[commandParams[i].name] = arg;
        }
        /**
         * Handle Bidi calls
         */
        if (this.sessionId && BIDI_COMMANDS.includes(command)) {
            if (!this.eventMiddleware) {
                throw new Error('Your WebDriver session doesn\'t support WebDriver Bidi');
            }
            return this.eventMiddleware[command](body.params);
        }
        const request = await RequestFactory.getInstance(method, endpoint, body, isHubCommand);
        request.on('performance', (...args) => this.emit('request.performance', ...args));
        this.emit('command', { method, endpoint, body });
        log.info('COMMAND', commandCallStructure(command, args));
        return request.makeRequest(this.options, this.sessionId).then((result) => {
            if (result.value != null) {
                let resultLog = result.value;
                if (/screenshot|recording/i.test(command) && typeof result.value === 'string' && result.value.length > 64) {
                    resultLog = `${result.value.slice(0, 61)}...`;
                }
                else if (command === 'executeScript' && body.script && body.script.includes('(() => window.__wdioEvents__)')) {
                    resultLog = `[${result.value.length} framework events captured]`;
                }
                log.info('RESULT', resultLog);
            }
            this.emit('result', { method, endpoint, body, result });
            if (command === 'deleteSession' && !process.env.WDIO_WORKER_ID) {
                logger.clearLogger();
            }
            return result.value;
        });
    };
}
